package util

import (
	"fmt"
	"os"
	"os/user"
	"path/filepath"
	"runtime"
	"strings"
	"unicode/utf8"
)

func ExpandPath(path string) string {
	var result string

	if strings.HasPrefix(path, "~") {
		usr, err := user.Current()
		if err != nil {
			panic(err)
		}

		dir := usr.HomeDir

		if path == "~" {
			result = dir
		} else {
			result = filepath.Join(dir, path[2:])
		}
	} else {
		var err error
		result, err = filepath.Abs(path)
		if err != nil {
			panic(err) // stupid interface of filepath.Abs
		}
	}

	return result
}

func SplitFilepath(path string) (result []string) {
	separator := os.PathSeparator

	path = filepath.FromSlash(path)
	path = filepath.Clean(path)

	if runtime.GOOS == "windows" {
		sepStr := string(separator)
		uncRootPath := fmt.Sprintf("%s%s", sepStr, sepStr)

		path = strings.TrimPrefix(path, uncRootPath)
		path = strings.TrimPrefix(path, sepStr)
		path = strings.TrimSuffix(path, sepStr)
	} else if filepath.IsAbs(path) {
		p, err := filepath.Rel(string(separator), path)
		if err != nil {
			panic(fmt.Sprintf("unable to get relative path for %q", path))
		}
		path = p
	}

	if path == "" || path == "." || path == string(separator) {
		return nil
	}

	idx := 0
	if separator == '\\' {
		// if the separator is '\\', then we can just split...
		result = strings.Split(path, string(separator))
		idx = len(result)
	} else {
		// otherwise, we need to be careful of situations where the separator was escaped
		cnt := strings.Count(path, string(separator))
		if cnt == 0 {
			return []string{path}
		}

		result = make([]string, cnt+1)
		pathlen := len(path)
		separatorLen := utf8.RuneLen(separator)
		emptyEnd := false
		for start := 0; start < pathlen; {
			end := indexRuneWithEscaping(path[start:], separator)
			if end == -1 {
				emptyEnd = false
				end = pathlen
			} else {
				emptyEnd = true
				end += start
			}
			result[idx] = path[start:end]
			start = end + separatorLen
			idx++
		}

		// If the last rune is a path separator, we need to append an empty string to
		// represent the last, empty path component. By default, the strings from
		// make([]string, ...) will be empty, so we just need to increment the count
		if emptyEnd {
			idx++
		}
	}

	return result[:idx]
}

// Find the first index of a rune in a string,
// ignoring any times the rune is escaped using "\".
func indexRuneWithEscaping(s string, r rune) int {
	end := strings.IndexRune(s, r)
	if end == -1 {
		return -1
	}
	if end > 0 && s[end-1] == '\\' {
		start := end + utf8.RuneLen(r)
		end = indexRuneWithEscaping(s[start:], r)
		if end != -1 {
			end += start
		}
	}
	return end
}

// GlobPrefixWithoutPatterns figures out how many components we don't need to glob because they're
// just names without patterns
func GlobPrefixWithoutPatterns(glob string) (string, string) {
	var prefix string

	globComponents := SplitFilepath(glob)
	length := len(globComponents)
	ind := 0
	for ; ind < length; ind++ {
		if strings.ContainsAny(globComponents[ind], "*?[{\\") {
			break
		}
	}
	if ind > 0 {
		prefix = filepath.Join(globComponents[0:ind]...)
		glob = filepath.Join(globComponents[ind:]...)
	}

	return prefix, glob
}

func FilepathsWithParents(path string) []string {
	var res []string
	base := ""
	for _, part := range SplitFilepath(path) {
		base = filepath.Join(base, part)
		res = append(res, base)
	}

	return res
}
